#include "EvolutionController.h"

EvolutionController::EvolutionController(ODEWorld* theWorld, Timer* theTimer, 
										 point3d initialHumanPosition)
{
	world = theWorld;
	timer = theTimer;
	countDown = 0;
	distanceTraveled = 0;
	initialPosition = initialHumanPosition;
	finalPosition.x = 0;
	finalPosition.y = 0;
	finalPosition.z = 0;
	generationNumber=0;
	fitnessType=-1;
	timeScale=-1;
	//brainSize[2] = {1, 8}; 
	brainSize[0] = 2;
	brainSize[1] = 5;
	brainSize[2] = 5;
	populationSize = 0;
	useInputFile = ' ';
	averageHeight = 0;
	numberOfCycles = 0;
	cutoffHeight = 0;
	highestHeight = 0;
	numberOfStepsTaken = 0;
	previousLeftFootHeight = 0;
	previousRightFootHeight = 0;
	steppingThreshold = 1.0;
	lastFootThatStepped = 'x'; //not initialized to either foot
}

EvolutionController::~EvolutionController()
{
	while (!brainList.empty())
	{
		delete brainList.back();
		brainList.pop_back();
	}
}

void EvolutionController::EvolutionLoop(ODEHuman* humanPtr, float deltaTime)
{
	//update countDown value
	countDown -= deltaTime;

	if (humanPtr != NULL)
	{
		//fitness function-specific calls go here

		//these values get reset in fitnessFunction
		averageHeight += humanPtr->GetHeadPosition().y;
		numberOfCycles++;
		
		if (humanPtr->GetHeadPosition().y > highestHeight)
			highestHeight = humanPtr->GetHeadPosition().y;

		//Now checking for steps (alternating steps only!)

		//check for left foot steps
		if (humanPtr->LeftFootMovingDownward())
		{
			if (previousLeftFootHeight > steppingThreshold && 
				humanPtr->LeftFootHeight() < steppingThreshold &&
				lastFootThatStepped != 'l')
			{
				//foot just moved down past threshold
				numberOfStepsTaken++;
				lastFootThatStepped = 'l';
			}
		}

		//check for right foot steps
		if (humanPtr->RightFootMovingDownward())
		{
			if (previousRightFootHeight > steppingThreshold && 
				humanPtr->RightFootHeight() < steppingThreshold &&
				lastFootThatStepped != 'r')
			{
				//foot just moved down past threshold
				numberOfStepsTaken++;
				lastFootThatStepped = 'r';
			}
		}

		//setting variables
		previousLeftFootHeight = humanPtr->LeftFootHeight();
		previousRightFootHeight = humanPtr->RightFootHeight();
	}

	if (humanPtr != NULL)
	{
		//end trial if time runs out OR person head gets too low (OR body is not upright - walking only)
		if (countDown <= 0 || humanPtr->GetHeadPosition().y < cutoffHeight
			|| (fitnessType == 2 && 
			fabs(humanPtr->GetHeadPosition().z) > fabs(humanPtr->GetPosition().z) + UPRIGHTNESS_FACTOR))
		{
			finalPosition = humanPtr->GetPosition();

			float fitness = FitnessFunction(humanPtr, initialPosition, finalPosition, fitnessType);

			(*brainIter)->SetFitness(fitness);
			std::cout << "Fitness = " << fitness << std::endl;

			//now try next brain...
			//delete humanPtr;
			//humanPtr = new ODEHuman(world, initialPosition.x, initialPosition.y, initialPosition.z, BODY_SIZE);
			humanPtr->SetInitialPosition(initialPosition);
			brainIter++;
			if (brainIter != brainList.end())
			{
				humanPtr->SetBrain(*brainIter);
			}
			//initialPosition = humanPtr->GetPosition();

			//reset variables
			countDown = EVOLUTION_TIME; 
		}
	}

	if (brainIter == brainList.end()) //end of generation
	{
		float averageFitness=0;
		std::cout << "End of Generation " << generationNumber;
		
		//1. sort brain list by fitness
		brainList.sort(ComparisonFunction); //highest value at the front of the list

		//2. save stats, best brain
		for (brainIter = brainList.begin(); brainIter != brainList.end(); brainIter++)
		{
			averageFitness += (*brainIter)->GetFitness();
		}
		averageFitness = averageFitness/populationSize;

		std::cout << ". Best: " << (*brainList.begin())->GetFitness() << ". Average: " << averageFitness << std::endl;

		//only write to files every 10 generations
		if (generationNumber != 0 && generationNumber%FILE_WRITING_PERIOD == 0)
		{
			//write to stats file
			statsFile = fopen("stats/stats.dat", "a");
			//output format: generation#,best,average; this format is good for Microsoft Excel importing
			fprintf(statsFile, "%d,%f,%f\n", generationNumber, (*brainList.begin())->GetFitness(), averageFitness);
			fclose(statsFile);

			//write to genes files
			//char buffer[32];
			std::string filename = "genes/g";
			//filename += itoa(generationNumber, buffer, 10); //this uses VC++ version
			filename += itoa(generationNumber);
			filename += "_best.nn";
			genesFile = fopen(filename.c_str(), "w");
			(*brainList.begin())->PrintGenesToFile(genesFile);
			fclose(genesFile);

			//wait a little while before starting the next generation; the file writing sometimes
			//messes up the beginning of the next generation since it takes a while to write.
			std::cout << std::endl << "=====================================" << std::endl;
			std::cout << "Writing data to output files..." << std::endl;
			std::cout << "=====================================" << std::endl << std::endl;

			float waitValue = 4;
			while (waitValue >=0)
			{
				waitValue -= timer->GetElapsedSeconds();
			}
		}

		//3. remove worst half
		for (int i=0; i<populationSize/2; i++)
		{
			delete brainList.back();
			brainList.pop_back();
		}

		//4. create new genes by crossover and mutate them; add new brains to the list
		NNGeneString* newGenes1 = NULL;
		NNGeneString* newGenes2 = NULL;
		std::list<NeuralNet*>::iterator brainIter2;
		NeuralNet* brainPtr;

		brainIter = brainList.begin();
		brainIter2 = brainList.begin();

		for (int i=0; i<populationSize/4; i++)
		{
			//Getting number between 0 and population/2-1, inclusive.
			//This number will be the number of times to increment brainIter2.
			int randomNum = rand()%(populationSize/2);

			for (int j=0; j<randomNum; j++)
			{
				brainIter2++; //randomly choosing mate for brainIter
			}

			CrossoverAndMutation((*brainIter), (*brainIter2), newGenes1, newGenes2);
			brainPtr = new NeuralNet(newGenes1, 0); //creating 1st new brain
			brainList.push_back(brainPtr);
			brainPtr = new NeuralNet(newGenes2, 0); //creating 2nd new brain
			brainList.push_back(brainPtr);
			brainIter++; //move to next brain
			brainIter2 = brainList.begin(); //reset this iterator
			delete newGenes1;
			delete newGenes2;
		}

		//Old, less random way...

		//brainIter = brainList.begin();
		//brainIter2 = brainIter;
		//for (int i=0; i<populationSize/2-1; i++)
		//{
		//	brainIter2++;
		//}

		////now brainIter points to item 0, brainIter2 points to list size-1

		//for (int i=0; i<populationSize/4; i++)
		//{
		//	CrossoverAndMutation((*brainIter), (*brainIter2), newGenes1, newGenes2);
		//	brainPtr = new NeuralNet(newGenes1, 0); //creating 1st new brain
		//	brainList.push_back(brainPtr);
		//	brainPtr = new NeuralNet(newGenes2, 0); //creating 2nd new brain
		//	brainList.push_back(brainPtr);
		//	brainIter++;
		//	brainIter2--;
		//	delete newGenes1;
		//	delete newGenes2;
		//}

		//reset things
		brainIter = brainList.begin();
		humanPtr->SetBrain(*brainIter);
		generationNumber++;

		timer->GetElapsedSeconds(); //reset timer before continuing
	}
}

float EvolutionController::FitnessFunction(ODEHuman* humanPtr, point3d start, 
										   point3d finish, int function)
{
	switch(function)
	{
		case 1: //average height - distance between feet
			{
				float fitness = averageHeight/numberOfCycles;

				//penalty for sliding feet apart
				fitness -= humanPtr->ZDistanceBetweenFeet();

				//penalty for falling below threshold height; penalty is greater for falling sooner
				if (countDown > 0)
				{
					fitness = fitness / countDown;
				}

				//reset average height stuff
				averageHeight = 0;
				numberOfCycles = 0;

				return fitness;
			}
			break;
		//case 2: //forward distance of torso * (forward distance of feet - height of feet) * average height
		//	{
		//		float fitness;
		//		float torsoDistance = finish.z - start.z;
		//		float footBonus = humanPtr->ForwardDistanceOfFeet() - humanPtr->CombinedFeetHeight();
		//		float myAverageHeight = averageHeight/numberOfCycles;

		//		if (torsoDistance < 0 && footBonus < 0)
		//		{
		//			fitness = torsoDistance * (-footBonus) * myAverageHeight;
		//		}
		//		else
		//		{
		//			fitness = torsoDistance * footBonus * myAverageHeight;
		//		}

		//		//penalty for falling below threshold height; penalty is greater for falling sooner
		//		if (countDown > 0)
		//		{
		//			fitness = fitness / countDown;
		//		}

		//		//reset average height stuff
		//		averageHeight = 0;
		//		numberOfCycles = 0;

		//		return fitness;
		//	}
		//	break;
		case 2: //torso distance * feet distance * (# of steps) * (average height)
			{
				float fitness;
				float feetDistance = humanPtr->ForwardDistanceOfFeet();
				float torsoDistance = finish.z - start.z;
				float myAverageHeight = averageHeight/numberOfCycles;

				if (torsoDistance < 0 && feetDistance < 0)
				{
					fitness = torsoDistance * (-feetDistance) * numberOfStepsTaken * myAverageHeight;
				}
				else
				{
					fitness = torsoDistance * feetDistance * numberOfStepsTaken * myAverageHeight;
				}

				//fitness = torsoDistance * numberOfStepsTaken * myAverageHeight;

				//penalty for falling below threshold height; penalty is greater for falling sooner
				if (countDown > 0)
				{
					fitness = fitness / countDown;
				}

				//reset variables
				previousLeftFootHeight = 0;
				previousRightFootHeight = 0;
				numberOfStepsTaken = 0;
				averageHeight = 0;
				numberOfCycles = 0;
				lastFootThatStepped = 'x'; //not initialized to either foot

				return fitness;
			}
			break;
		case 3: //highest head height
			{
				float fitness = highestHeight;
				highestHeight = 0;

				//note: no penalty for falling early

				return fitness;
			}
			break;
		//case 4: //number of foot steps * average height
		//	{
		//		float fitness = (leftFootSteps + rightFootSteps)*(averageHeight/numberOfCycles);

		//		//penalty for falling below threshold height; penalty is greater for falling sooner
		//		if (countDown > 0)
		//		{
		//			fitness = fitness / countDown;
		//		}

		//		//reset variables
		//		previousLeftFootHeight = 0;
		//		previousRightFootHeight = 0;
		//		leftFootSteps = 0;
		//		rightFootSteps = 0;
		//		averageHeight = 0;
		//		numberOfCycles = 0;
		//		return fitness;
		//	}
		//	break;

		//case 4: //2D distance (XZ plane)
		//	return sqrt(pow((finish.x - start.x),2) + pow((finish.z - start.z),2));
		//	break;
		//case 10: //3D distance
		//	return sqrt(pow((finish.x - start.x),2) + pow((finish.y - start.y),2) 
		//		+ pow((finish.z - start.z),2));
		//	break;
		default:
			break;
	}
}

bool ComparisonFunction(NeuralNet* a, NeuralNet* b)
{
	return (a->GetFitness() > b->GetFitness());
}

NeuralNet* EvolutionController::GetFirstBrain()
{
	return (*brainList.begin());
}

void EvolutionController::GetUserInput()
{
	std::cout << "Welcome to the Virtual Human Evolver!" << std::endl;
	std::cout << "=====================================" << std::endl << std::endl;
	std::cout << "When the graphical window appears," << std::endl;
	std::cout << "drag the mouse buttons over it to adjust the view." << std::endl;
	std::cout << "Press 'q' in the graphical window to quit." << std::endl << std::endl;

	while (useInputFile != 'y' && useInputFile != 'n')
	{
		std::cout << "Would you like to use a saved neural network file\n(must be named input.nn in genes directory)? ";
		std::cin >> useInputFile;
		if (useInputFile != 'y' && useInputFile != 'n')
			std::cout << "Please type y or n" << std::endl;
	}

	while (populationSize == 0)
	{
		std::cout << std::endl << "Enter the population size (must be a multiple of 4): ";
		std::cin >> populationSize;
		if (populationSize%4 != 0 || populationSize == 0)
		{
			std::cout << "That was not a multiple of 4.  Try again." << std::endl;
			populationSize = 0;
		}
	}

	while (fitnessType == -1)
	{
		std::cout << std::endl << "=====================================" << std::endl << std::endl;
		std::cout << "1. Standing (fitness = average height - distance between feet)" << std::endl;
		//std::cout << "2. Walking Forward (fitness = torso distance * (feet distance " << std::endl << "   - distance between feet) * average height)" << std::endl;
		std::cout << "2. Walking (fitness = torso distance * feet distance *" << std::endl << "     # of steps * average height)" << std::endl;
		std::cout << "3. Jumping (fitness = highest height of head)" << std::endl << std::endl;
		//std::cout << "4. Stepping (fitness = # of foot movements downward * average height)" << std::endl << std::endl;
		std::cout << "Which type of action do you want to evolve? ";
		std::cin >> fitnessType;
		if (fitnessType < 1 || fitnessType > 3)
		{
			std::cout << "Invalid entry." << std::endl << std::endl;
			fitnessType = -1;
		}
	}

	if (fitnessType == 1 || fitnessType == 2) //standing or walking
	{
		cutoffHeight = CUTOFF_HEIGHT_1;
	}
	else if (fitnessType == 3) //jumping
		cutoffHeight = CUTOFF_HEIGHT_2;

	if (useInputFile == 'y') //creating the first batch of brains (half will be clones of input file)
	{
		genesFile = fopen("genes/input.nn", "r");
		if (genesFile == NULL)
		{
			std::cout << "Invalid input file!  ./genes/input.nn does not exist.  Quitting..." << std::endl;
			exit(1);
		}

		std::cout << std::endl << "One will be a clone of input.nn, the rest will be random." << std::endl << std::endl;
		
		//cloning input file
		NeuralNet* brainPtr;
		brainPtr = new NeuralNet(NUM_BRAIN_INPUTS, brainSize, NUM_BRAIN_OUTPUTS, genesFile, ACTIVATION_FUNCTION);
		brainList.push_back(brainPtr);
		fclose(genesFile);
		genesFile = fopen("genes/input.nn", "r");

		fclose(genesFile);
		for (int i=1; i<populationSize; i++) //creating random brains
		{
			brainPtr = new NeuralNet(NUM_BRAIN_INPUTS, brainSize, NUM_BRAIN_OUTPUTS, ACTIVATION_FUNCTION);
			brainList.push_back(brainPtr);
		}
	}
	else if (useInputFile == 'n') //creating the first batch of brains (initially random)
	{
		NeuralNet* brainPtr;
		for (int i=0; i<populationSize; i++)
		{
			brainPtr = new NeuralNet(NUM_BRAIN_INPUTS, brainSize, NUM_BRAIN_OUTPUTS, ACTIVATION_FUNCTION);
			brainList.push_back(brainPtr);
		}
	}

	brainIter = brainList.begin();
	countDown = EVOLUTION_TIME;

	//while (timeScale == -1)
	//{
	//	std::cout << "Please enter a time scale from 1 to 10 (1=real-time, 2=double time, etc.): ";
	//	std::cin >> timeScale;
	//	if (timeScale < 1 || timeScale > 10)
	//	{
	//		std::cout << "Invalid entry." << std::endl << std::endl;
	//		timeScale = -1;
	//	}
	//}

	std::cout << std::endl << "The file ./stats/stats.dat will be appended." << std::endl;
	std::cout << "*.nn files in ./genes will be created/overwritten." << std::endl << std::endl;
	std::cout << "=====================================" << std::endl << std::endl;

	//opening stats output file
	statsFile = fopen("stats/stats.dat", "a");
	fprintf(statsFile, "New Run Started: Task %d, Population Size %d, ", fitnessType, populationSize);
	if (useInputFile == 'y')
	{
		fprintf(statsFile, "Using input neural net file to seed population\n");
	}
	else if (useInputFile == 'n')
	{
		fprintf(statsFile, "No input file - totally random initial population\n");
	}
	fclose(statsFile);
}

/*--------------------------------------------------------------------+
| itoa() - manage the sign, compute the string equivalent, and call   |
| memcpy().                                                           |
+--------------------------------------------------------------------*/
char *itoa(int value)
{
	int count,                   /* number of characters in string       */
	    i,                       /* loop control variable                */
	    sign;                    /* determine if the value is negative   */
	char *ptr,                   /* temporary pointer, index into string */
	     *string,                /* return value                         */
	     *temp;                  /* temporary string array               */
	
	count = 0;
	if ((sign = value) < 0)      /* assign value to sign, if negative    */
	   {                         /* keep track and invert value          */
	   value = -value;
	   count++;                  /* increment count                      */
	   }
	
	/* allocate INTSIZE plus 2 bytes (sign and NULL)                     */
	temp = (char *) malloc(INTSIZE + 2);
	if (temp == NULL)
	   {
	   return(NULL);
	   }
	memset(temp,'\0', INTSIZE + 2);
	
	string = (char *) malloc(INTSIZE + 2);
	if (string == NULL)
	   {
	   return(NULL);
	   }
	memset(string,'\0', INTSIZE + 2);
	ptr = string;                /* set temporary ptr to string          */
	
	/*--------------------------------------------------------------------+
	| NOTE: This process reverses the order of an integer, ie:            |
	|       value = -1234 equates to: char [4321-]                        |
	|       Reorder the values using for {} loop below                    |
	+--------------------------------------------------------------------*/
	do {
		   *temp++ = value % 10 + '0';   /* obtain modulus and or with '0'   */
	   count++;                      /* increment count, track iterations*/
	   }  while (( value /= 10) >0);
	
	if (sign < 0)                /* add '-' when sign is negative        */
	   *temp++ = '-';
	
	*temp-- = '\0';              /* ensure null terminated and point     */
	                             /* to last char in array                */
	
	/*--------------------------------------------------------------------+
	| reorder the resulting char *string:                                 |
	| temp 	- points to the last char in the temporary array               |
	| ptr  - points to the first element in the string array              |
	+--------------------------------------------------------------------*/
	for (i = 0; i < count; i++, temp--, ptr++)
	   {
	   memcpy(ptr,temp,sizeof(char));
	   }
	
	return(string);
}

